#!/bin/perl
use IO::Select;
use Switch;
use Socket;
use APRSdecu;
# use strict;

my $aprsport = "localhost:2023"; #servername and port number for Xastir server port
my $user = "KI4HDU-4"; #callsign to log into Xastir with
my $pass = "18614";    #callpass to log into Xastir with
my $locations = "locations.dat" ;
my $stations = "stations.dat";
my $mi2nm  = 0.8689762;
my %locations;
my %stations;
my %aliases;
my $directions = "\nENTER STATION LOCATION (or 'q' to quit):\nSTATION ID, WAYPOINT ID, SPEED, COURSE, COMMENTS\n";
my $displaypackets = 0;

# ========================================================================
#                    Supporting Functions
# ========================================================================
sub OpenAPRSPORT(){
	print "Connecting to Xastir server...";
	my ($host, $port) = split/:/, $aprsport, 2;
	my $aprs_addr = (gethostbyname($host)) [4];
	my $addr = sockaddr_in($port, $aprs_addr);
	my $proto = getprotobyname('tcp');
	
	# Create an Internet protocol socket.
	socket(APRSPORT, AF_INET, SOCK_STREAM, $proto) or die "socket:$!";
	
	# Connect our socket to the server socket
	connect (APRSPORT, $addr) or die "connect:$!";
	
	# For fflish on socket file handle after every write
	select (APRSPORT);
	$| = 1;
	select(STDOUT);
	
	printf APRSPORT "user %s pass %s\n\r", $user, $pass;
	printf "done!\nConnected as user %s wth pass %s\n\n", $user, $pass;
}

sub executeCommand(*) {
	my $validCMD = 1;
	my ($cmd, $args ) = split / /, $_[0] , 2;
	switch($cmd) {
		case "q" { exit; }
		case "quit" { exit; }
		case "sign" { UploadSignpostObjects() ; }
		case "reload" {
				%locations = ();
				ReadLocations(); 
				%stations = ();
				%aliases = ();
				ReadStations();
			}
		case "lstations" { ListStations( $args ); }
		case "llocations" { ListLocations( $args ); }
		case "ustations" { ReadStations(); }
		case "ulocations" { ReadLocations(); }
		case "ls" { ListStations( $args ); }
		case "ll" { ListLocations( $args ); }
		case "la" { ListAliases( $args ); }
		case "help" { Help() ; }
		case "msg" { SendMessage( $args ); }
		case "message" { SendMessage( $args ); }
		case "nodisplay" { $displaypackets = 0; }
		case "display" { $displaypackets = 1; }
		else { $validCMD = 0; }
	}
	return $validCMD;
}

sub Help() {
print <<EOF;
----------------------------------------------------------------------
-                              H E L P                               -
----------------------------------------------------------------------
q                           Quit
sign                        UploadSignpostObjects
reload                      Clear and Reload Locations and Stations
ls - lstations [station]    List Stations
ll - llocations [location]  List Locations
ustations                   Reload Stations (do not clear)
ulocations                  Reload Locations (do not clear)
display                     Turn on packet display
nodisplay                   Turn off packet display
message STATION MESSAGE     Send MESSAGE to STATION
----------------------------------------------------------------------
EOF
print $directions;
}

sub isValidLocation(*) {
	my $totest = $_[0];
	
	if( $locations{$totest}{"lat"} eq "" ) {
		return 0;
	} else {
		return 1;
	}
}

sub isValidStation(*) {
	my $totest = $_[0];
	
	if( $stations{$totest}{"pri"} eq "" ) {
		return 0 ;
	} else {
		return 1;
	}
}

sub ReadLocations( ){
	print "Reading locations...";
	open(LOCATIONS, "<$locations") or die("Couldn't open $locations\n") ;
	my $line = 0;
	while(<LOCATIONS>){
		$line++;
		chomp;
		if( $_ =~ m/^#/ ) { next; }
		my ($loc, $lat, $lon ) = split /,/;
		if( $lat !~ m/\d\d\d\d\.\d\d[NS]/ || $lon !~ m/\d\d\d\d\d\.\d\d[EW]/ ){ 
			print "\nProblem with $locations on line: $line - $loc.\n";
			next;
		}
		$locations{ $loc }{ "lat" } = $lat;
		$locations{ $loc }{ "lon" } = $lon;
	
		$loc = uc($loc);
		$locations{ $loc }{ "lat" } = $lat;
		$locations{ $loc }{ "lon" } = $lon;
	
		$loc = lc($loc);
		$locations{ $loc }{ "lat" } = $lat;
		$locations{ $loc }{ "lon" } = $lon;
	}
	print "done!\n";
}

sub ReadStations( ){
	print "Reading stations...";
	open(STATIONS, "<$stations") or die("Couldn't open $stations\n") ;
	my $line = 0;
	while(<STATIONS>){
		$line++;
		chomp;
		my ($instation, $symbol, $aliases ) = split /,/;
		my @symbol = split  //, $symbol ;
		if( length($instation) > 9 ) {
			print "\nProblem with $stations on line: $line - $instation name too long.\n";
			next;
		}
#		printf "%s is %s characters long\n", $instation, length($instation);
		$stations{$instation}{ "pri" } = $symbol[0];
		$stations{$instation}{ "sec" } = $symbol[1];
		$stations{$instation}{ "aliases" } = $aliases ;
		$aliases{$instation} = $instation;
		$aliases{uc($instation)} = $instation;
		$aliases{lc($instation)} = $instation;
		foreach my $alias ( split / /, $aliases ) {
			$aliases{$alias} = $instation;
			$aliases{uc($alias)} = $instation;
			$aliases{lc($alias)} = $instation;
		}
	}
	print "done!\n";
}

sub UploadSignpostObjects( ) {
	print "Uploading Signpost Objects for Waypoints...\n";
	foreach my $loc( sort keys %locations) {
		printf APRSPORT "%s>APZWJM:)%s!%s\\%sm {%s}\n", $user, $loc, $locations{$loc}{"lat"},$locations{$loc}{"lon"},$loc;
		printf "%s>APZWJM:)%s!%s\\%sm {%s}\n", $user, $loc, $locations{$loc}{"lat"},$locations{$loc}{"lon"},$loc;
	}
	print "done!\n";
}

sub ListAliases( ) {
	foreach $al ( sort keys %aliases ) {
		printf "%s = %s\n", $al, $aliases{$al};
	}
}
sub ListLocations( ){
	my $loc;
	if( $_[0] eq "" ) {
		foreach $loc ( sort keys %locations ) {
			printf "%s  at %s %s\n", $loc, $locations{$loc}{"lat"},$locations{$loc}{"lon"};
		}
	} else {
		$loc = $_[0];
		printf "%s  at %s %s\n", $loc, $locations{$loc}{"lat"},$locations{$loc}{"lon"};
	}
}

sub ListStations( ){
	my $sta;
	my $alias;
	if( $_[0] eq "" ) {
		foreach $sta ( sort keys %stations ) {
			printf "%s %s%s\n", $sta, $stations{$sta}{"pri"},$stations{$sta}{"sec"};
			foreach $alias(split / /, $stations{$sta}{"aliases"} ) {
				printf "\t%s\n", $alias;
			}
		}
	} else {
		$sta = $aliases{ $_[0] } ;
		printf "%s %s%s\n", $sta, $stations{$sta}{"pri"},$stations{$sta}{"sec"};
		foreach $alias(split / /, $stations{$sta}{"aliases"} ) {
			printf "\t%s\n", $alias;
		}
	}
}

sub SendMessage( ){
	my ( $to, $msg ) = split / /, $_[0], 2 ;
	my $tostation = $aliases{ $to } ;
	if( $tostation eq "" ) { $tostation = $to; }
	printf APRSPORT "%s>APZWJM::%-9.9s:%.67s{%3.3dO}\n", $user, $tostation, $msg, $messagecounter;
	printf "%s>APZWJM::%-9.9s:%.67s{%3.3dO}\n", $user, $tostation, $msg, $messagecounter;
	$messagecounter++;
}

sub UserInput(*){
	$inline = shift;
	chomp($inline);
	if( $inline ne "" ) {
		if( executeCommand( $inline ) ) { next; }
		my ( $station, $location, $speed, $course, $comment ) = split /,/, $inline;
		if( $speed eq "" ) { $speed = "000"; } else { $speed = int( $speed * $mi2nm + 0.5 ) }
		if( $course eq "" ) { $course = "000"; }
		print "Station = $station\n" ;
		$station = $aliases{ $station } ;
		if(  isValidStation($station) ) {
			if( isValidLocation($location) ) {
				my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time);
				$heardstations{$station}{ "location" } = $location;
				$heardstations{$station}{ "time" } = $hour . ":" . $min . ":" . $sec;
				printf APRSPORT "%s>APZWJM,TCPIP:;%-9.9s*%.2d%.2d%.2dz%s%s%s%s%.3d/%.3d%s\n", $user, $station, $mday, $hour, $min, $locations{ $location }{"lat"}, $stations{$station}{"pri"}, $locations{ $location }{"lon"}, $stations{$station}{"sec"}, $course, $speed, $comment;
				printf "%s>APZWJM,TCPIP:;%-9.9s*%.2d%.2d%.2dz%s%s%s%s%.3d/%.3d%s\n", $user, $station, $mday, $hour, $min, $locations{ $location }{"lat"}, $stations{$station}{"pri"}, $locations{ $location }{"lon"}, $stations{$station}{"sec"}, $course, $speed, $comment;
			} else {
				printf "Invalid location code: %s\n", $location;
			}
		} else {
			printf "Invalid command or station ID: %s\n", $station;
		}
	}
}


sub Usage(){
	print "\nUsage:\n";
	print "\ttracker [-h | -help]  [-v | -version]   [-u | -user callsign]\n";
	print "\t\t [-p | -pass callpass] [-port hostname:port]\n\n";
	print "\t-h or -help             for this help screen\n";
	print "\t-v or -version          to display version\n";
	print "\t-u or -user             callsign to log into Xastir\n";
	print "\t-p or -pass             callpass to log into Xastir\n";
	print "\t-locations              filename for locations/waypoints\n";
	print "\t-stations               filename for stations\n";
	print "\t-port                   computer Xastir server host:port\n\n";
	exit(0);
}

sub Version(){
	print "\nName:    tracker\n";
	print "Version: Version 0.1\n";
	print "Author:  William McKeehan, KI4HDU\n\n";
	print "This program may be copied only under the terms of\n";
	print "the GNU General Public License.\n\n";
	exit(0);
}

# ========================================================================
#                    COMMAND LINE ARGUMENT PARSING
#             Parse the input line for command line settings
# ========================================================================

for (my $i = 0; $i <= $#ARGV; $i++) {
    switch ( $ARGV[$i] ) {
        case "-u" { $i++; $user = $ARGV[$i]; }
		case "-user" { $i++; $user = $ARGV[$i]; }
        case "-p" { $i++; $pass = $ARGV[$i]; }
		case "-pass" { $i++; $pass = $ARGV[$i]; }
        case "-port" { $i++; $aprsport = $ARGV[$i]; }
		case "-locations" { $i++; $locations = $ARGV[$i]; }
		case "-stations" { $i++; $stations = $ARGV[$i]; }
        case "-help" { Usage(); }
        case "-h" { Usage(); }
        case "-v" { Version(); }
        case "-version" { Version(); }
        else { die "\t*** Invalid argument = $ARGV[$i]\n"; }
    }
}

# ========================================================================
#                                  MAIN
# ========================================================================
OpenAPRSPORT();
ReadLocations();
ReadStations();
#UploadSignpostObjects(); #enter the command "sign" if you want this to happen.

# Initalize
my $aprsdecu = new APRSdecu;
my $read_set = new IO::Select();
$read_set->add(\*STDIN);
$read_set->add(\*APRSPORT);

my $counter = 0;
my $messagecounter = 0;

print $directions ;

while(1){
	my @rh_set = $read_set->can_read(0);
	foreach $rh (@rh_set){
		$buf = <$rh>;
		if( $buf) {
			if( $rh == \*STDIN ){ UserInput($buf); $counter++; }
			else { if( $displaypackets ) { $aprsdecu->DisplayPacket($buf); }}
		}
		if( $counter == 100 ) { print $directions ; $counter = 0; }
		if( (int($counter)/10)==int(int($counter)/10) ) { print APRSPORT "#WJM "; }
	}
}
